/* d_coil.c  94.03.12
 * Copyright 1983-1992   Albert Davis
 * inductors
 * two levels: linear (lin) and nonlinear (nl) (not really)
 * x = amps, y.f0 = flux, ev = y.f1 = henrys
 */
#include "ecah.h"
#include "ac.h"
#include "branch.h"
#include "error.h"
#include "mode.h"
#include "options.h"
#include "declare.h"
/*--------------------------------------------------------------------------*/
static	void	expand_coil(branch_t*);
static 	int	tr_coil_lin(branch_t*);
static 	int	tr_coil_nl(branch_t*);
static	void	integrate(branch_t*);
static 	void	ac_coil_lin(branch_t*);
static 	void	ac_coil_nl(branch_t*);
static	double	tr_review_coil(branch_t*);
/*--------------------------------------------------------------------------*/
extern const int stiff;		/* flag: use "stiff" integration method	    */
extern const double init_curr;	/* initial current			    */
extern const double init_volt;	/* initial voltage			    */
extern const double trtime0;	/* transient analysis time		    */
extern const int sim_mode;	/* what type simulation is this		    */
extern const int sim_phase;	/* how far it got			    */
extern const ac_t ac;
extern const struct options opt;
extern int in_curr;		/* flag: initial current		    */
extern int in_volt;		/* flag: initial voltage		    */
/*--------------------------------------------------------------------------*/
functions_t dev_coil = {
   (generic_t*)NULL,	/* x */
   sizeof(functions_t),	/* ssize */
   sizeof(branch_t),	/* elementsize */
   (functions_t*)NULL,	/* super */
   2, 			/* numnodes */
   rnONEPORT,		/* refnode */
   rnONEPORT,		/* isdevice */
   create_std,		/* create */
   copy_std,		/* copy */
   parse_std,		/* parse */
   print_std,		/* print */
   expand_coil,		/* expand */
   probe_std,		/* probe */
   tr_coil_nl,		/* dotr */
   unloadpassive,	/* untr */
   ac_coil_nl,		/* doac */
   trfix1,		/* trfun1 */
   trfix0,		/* trfun0 */
   acfix,		/* acfun */
   NULL,		/* tr_guess */
   NULL,		/* tr_advance */
   tr_review_coil	/* tr_review */
};
functions_t dev_coil_lin = {
   (generic_t*)NULL,	/* x */
   sizeof(functions_t),	/* ssize */
   sizeof(branch_t),	/* elementsize */
   &dev_coil,		/* super */
   2, 			/* numnodes */
   rnONEPORT,		/* refnode */
   rnONEPORT,		/* isdevice */
   create_std,		/* create */
   copy_std,		/* copy */
   parse_std,		/* parse */
   print_std,		/* print */
   expand_coil,		/* expand */
   probe_std,		/* probe */
   tr_coil_lin,		/* dotr */
   unloadpassive,	/* untr */
   ac_coil_lin,		/* doac */
   NULL,		/* trfun1 */
   NULL,		/* trfun0 */
   NULL,		/* acfun */
   NULL,		/* tr_guess */
   NULL,		/* tr_advance */
   tr_review_coil	/* tr_review */
};
/*--------------------------------------------------------------------------*/
static void expand_coil(branch_t *brh)
{
 if (!brh->x){
    brh->f = &dev_coil_lin;
    brh->y0.f1 = brh->val;
    brh->y0.f0 = LINEAR;
    brh->ev.x  = brh->val;
    brh->ev.y  = 0.;
 }
}
/*--------------------------------------------------------------------------*/
static int tr_coil_lin(branch_t *brh)
{
 trsetup(brh);
 brh->m0.x = tr_volts(&(brh->n[OUT1]),&(brh->n[OUT2]));
 brh->y0.x = brh->m0.c0 + brh->m0.f1 * brh->m0.x;
 brh->y0.f1 = brh->val;
 brh->y0.f0 = brh->y0.x * brh->y0.f1;
 if (brh->y0.f1 == 0.){
    error(bPICKY, "%s: short circuit\n", printlabel(brh,NO));
 }
 integrate(brh);
 trloadpassive(brh);
 return brh->converged = YES;
}
/*--------------------------------------------------------------------------*/
static int tr_coil_nl(branch_t *brh)
{
 in_volt = in_curr = NO;
 trsetup(brh);
 brh->m0.x = tr_volts_limited(&(brh->n[OUT1]),&(brh->n[OUT2]));
 brh->y0.x = brh->m0.c0 + brh->m0.f1 * brh->m0.x;
 if (brh->f->trfun1){
    (*brh->f->trfun1)(brh);
 }else{
    brh->y0.f1 = brh->val;
    brh->y0.f0 = brh->y0.x * brh->y0.f1;
 }
 if (brh->y0.f1 == 0.){
    error(bPICKY, "%s: short circuit\n", printlabel(brh,NO));
 }

 integrate(brh);
 trloadpassive(brh);
 return brh->converged = conv_check(brh);
}
/*--------------------------------------------------------------------------*/
static void integrate(branch_t *brh)
{
 if (sim_mode == sDC  ||  sim_mode == sOP  ||  brh->y0.f1 == 0.){
    brh->m0.f1 = 1./opt.shortckt;
    brh->m0.c0 = 0.;
 }else if (sim_phase == pINIT_DC){
    if (trtime0 == 0.){
       if (in_curr){
	  brh->m0.f1 = 0.;
	  brh->m0.c0 = -init_curr;
       }else if (in_volt){
	  brh->m0.f1 = 1./opt.shortckt;
	  brh->m0.c0 = -init_volt * brh->m0.f1;
       }else{
	  brh->m0.f1 = 1./opt.shortckt;
	  brh->m0.c0 = 0.;
       }
    }else{
       /* leave it alone to restore */
    }
 }else{
    double dt = brh->time0 - brh->time1;
    if (stiff){
       brh->m0.f1 = dt / brh->y0.f1;
       brh->m0.c0 = brh->mt1.c0 + brh->mt1.f1 * brh->mt1.x; /* oldi */
    }else{ /* not stiff */
       /* oldi = brh->mt1.c0 + brh->mt1.f1 * brh->mt1.x; */
       brh->m0.f1 = dt/(2*brh->y0.f1);
       brh->m0.c0 = brh->mt1.c0 + (brh->mt1.f1 + brh->m0.f1) * brh->mt1.x;
       		    /* oldi + f1*oldv, with combined terms */
    }
 }
}
/*--------------------------------------------------------------------------*/
static void ac_coil_lin(branch_t *brh)
{
 if (ac.omega == 0.){
    brh->acg.x = 1. / opt.shortckt;
    brh->acg.y = 0.;
    acloadpassivereal(brh);
    error(bPICKY, "%s: short circuit\n", printlabel(brh,NO));
 }else{
    brh->acg.x = 0.;
    brh->acg.y = -1. / (brh->ev.x * ac.omega);
    acloadpassiveimaginary(brh);
 }
}
/*--------------------------------------------------------------------------*/
static void ac_coil_nl(branch_t *brh)
{
 double dcvolts;

 if (brh->f->acfun){
    dcvolts = dc_volts(&(brh->n[OUT1]),&(brh->n[OUT2]));
    brh->acbias = brh->m0.c0 + brh->m0.f1*dcvolts;
    brh->ev = (*brh->f->acfun)(brh);
    if (ac.omega == 0.){
       brh->acg.x = 1. / opt.shortckt;
       brh->acg.y = 0.;
       error(bPICKY, "%s: short circuit\n", printlabel(brh,NO));
    }else{
       complex_t y;
       y.x = -brh->ev.y * ac.omega;
       y.y =  brh->ev.x * ac.omega;
       brh->acg = cflip(y);
    }
    acloadpassive(brh);
 }else{
    brh->ev.x  = brh->y0.f1;
    brh->ev.y  = 0.;
    if (ac.omega == 0.){
       brh->acg.x = 1. / opt.shortckt;
       brh->acg.y = 0.;
       acloadpassivereal(brh);
       error(bPICKY, "%s: short circuit\n", printlabel(brh,NO));
    }else{
       brh->acg.x = 0.;
       brh->acg.y = -1. / (brh->ev.x * ac.omega);
       acloadpassiveimaginary(brh);
    }
 }
}
/*--------------------------------------------------------------------------*/
static double tr_review_coil(branch_t *brh)
{
 if (brh->time2 <= 0.){
    return BIGBIG;
 }else{
    double factor = 12.;      /* 1/coefficient of 3rd der, trapezoid rule */
    double i0  = brh->m0.x;	/* volts */
    double it1 = brh->mt1.x;
    double it2 = brh->mt2.x;
    double didt0 = (i0  - it1) / (brh->time0 - brh->time1);
    double didt1 = (it1 - it2) / (brh->time1 - brh->time2);
    double ddiddt = (didt0 - didt1) / (brh->time0 - brh->time2);
    double q0  = brh->y0.f0;					/* flux */
    double qt1 = brh->yt1.f0;
    double dqdt0 = (q0  - qt1) / (brh->time0 - brh->time1);
    double currenttol = opt.abstol + opt.reltol * MAX(fabs(i0), fabs(it1));
    double chargetol = MAX(opt.chgtol,MAX(fabs(q0),fabs(qt1)))
			   * opt.reltol / (brh->time0 - brh->time1);
    double tol = MAX(currenttol,chargetol);
    double timestep =
	 opt.trtol * sqrt(factor * tol / MAX(opt.abstol, fabs(ddiddt)));

    if (opt.foooo == 943){
       printf("q0=%g v0=%g dq0=%g\n", q0, i0, dqdt0);
       printf("vt=%g qt=%g tol=%g\n", currenttol, chargetol, tol);
    }
    
    if (timestep < (brh->time0 - brh->time1) * opt.trreject){
       return brh->time1 + MAX(timestep,opt.mrt);
    }else{
       return brh->time0 + timestep;
    }
 }
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
